/*
 -------------------------------------------------------------------------------
 | Producer - Copyright (C) 2001-2007  Don Burns                               |
 |                                                                             |
 | This library is free software; you can redistribute it and/or modify        |
 | it under the terms of the GNU Lesser General Public License as published    |
 | by the Free Software Foundation; either version 3 of the License, or        |
 | (at your option) any later version.                                         |
 |                                                                             |
 | This library is distributed in the hope that it will be useful, but         |
 | WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  |
 | or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public     |
 | License for more details.                                                   |
 |                                                                             |
 | You should have received a copy of the GNU Lesser General Public License    |
 | along with this software; if not, write to the Free Software Foundation,    |
 | Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.               |
 |                                                                             |
 -------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>

#include <osgDB/ReadFile>
#include <osg/Node>
#include <osgUtil/SceneView>

#include <Producer/Camera.h>
#include <ProducerOSG/OsgSceneHandler.h>

#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Roller.H>

#include "FLTKRenderSurface.h"

static double viewAngle = 0.0;

static void redraw( Producer::Camera *camera )
{
    Producer::OsgSceneHandler *sh = dynamic_cast<Producer::OsgSceneHandler *>(camera->getSceneHandler() );

    Producer::Vec3 center; 
    double radius;
    sh->getSceneBounds( center, radius );

    Producer::Vec3 eye(  radius * 3 * cos(-viewAngle),
                         radius * 3 * sin(-viewAngle),
                         0.0 );

    Producer::Matrix viewMatrix;
    viewMatrix.makeLookAt( eye, center, Producer::Vec3( 0, 0, 1 ));

    camera->setViewByMatrix( viewMatrix );
    camera->frame();

    if( camera->getFrameCount() == 1 )
        camera->frame();
}

static void somecallback(Fl_Widget* w, void* v) 
{
}


static void quit_cb(Fl_Widget*, void*) 
{
    exit(0);
}


static Fl_Menu_Item menuitems[] = {
  { "&File",              0, 0, 0, FL_SUBMENU },
    { "&New",        0, (Fl_Callback *)somecallback },
    { "&Open...",    FL_CTRL + 'o', (Fl_Callback *)somecallback },
    { "&Save",       FL_CTRL + 's', (Fl_Callback *)somecallback },
    { "Save &As...", FL_CTRL + FL_SHIFT + 's', (Fl_Callback *)somecallback, 0, FL_MENU_DIVIDER },
    { "E&xit", FL_CTRL + 'q', (Fl_Callback *)quit_cb, 0 },
    { 0 },

  { "&Edit", 0, 0, 0, FL_SUBMENU },
    { "Cu&t",        FL_CTRL + 'x', (Fl_Callback *)somecallback },
    { "&Copy",       FL_CTRL + 'c', (Fl_Callback *)somecallback },
    { "&Paste",      FL_CTRL + 'v', (Fl_Callback *)somecallback },
    { "&Delete",     0, (Fl_Callback *)somecallback },
    { 0 },


  { "&Help", 0, 0, 0, FL_SUBMENU },
    { "&About...",       FL_CTRL + 'f', (Fl_Callback *)somecallback, 0, FL_MENU_DIVIDER },
    { "&Reference...",   FL_CTRL + 'f', (Fl_Callback *)somecallback },
    { "&Tutorial...",    FL_CTRL + 'f', (Fl_Callback *)somecallback },
    { 0 },

  { 0 }
};

static void draw_cb(Fl_Widget*, void* data) 
{
    Producer::Camera  *camera = static_cast<Producer::Camera *>(data);
    if( camera != NULL )
        redraw(camera);
}

static void roller_cb(Fl_Widget* w, void* data) 
{
    Fl_Roller *roller = static_cast<Fl_Roller *>(w);
    viewAngle = roller->value();

    Producer::Camera  *camera = static_cast<Producer::Camera *>(data);
    if( camera != NULL )
        redraw(camera);
}

int main(int argc, char **argv)
{
    // Load the database
    osg::ref_ptr<osg::Node> node =  osgDB::readNodeFile( "cow.osg" );
    if(  !node.valid() )
        return 1;

    /////////////////////////////////////////////////////////////////////
    // Initialize FLTK stuff
    Fl_Window *window = new Fl_Window( 660, 560, "FLTK Example" );
    Fl_Menu_Bar* m = new Fl_Menu_Bar(0, 0, 660, 30);
    m->copy(menuitems, window);

    // Create an FLTKRenderSurface which inherits from both 
    // Producer::RenderSurface and FLTK Fl_Gl_Window
    Producer::ref_ptr<FLTKRenderSurface> renderSurface = new FLTKRenderSurface( 10, 40, 640, 480);

    // Create the camera and set the FLTK RenderSurface to be its render surface
    Producer::ref_ptr<Producer::Camera> camera = new Producer::Camera;
    camera->setRenderSurface( renderSurface.get() );

    // We add a little functionality here a la FLTK, by adding a callback to redraw.
    renderSurface->setDrawCallback( draw_cb, camera.get() );

    // Set up the scene handler
    Producer::ref_ptr<Producer::OsgSceneHandler> osgSceneHandler = new Producer::OsgSceneHandler;
    if( node.valid() )
        osgSceneHandler->setScene( node.get() );
    camera->setSceneHandler( osgSceneHandler.get() );

    // Lets use a roller to rotate the view in a callback
    Fl_Roller *roller = new Fl_Roller( 270, 530, 120, 20 );
    roller->type(FL_HORIZONTAL);
    roller->minimum( -3.14159 * 0.5 );
    roller->maximum(  3.14159 * 0.5);
    roller->value(0.0);
    roller->callback( roller_cb, camera.get() );

    window->end();
    window->show(argc, argv);

    return Fl::run();
}

